#!/usr/bin/env python
# -*- coding: utf-8 -*-
__version__ = "0.3.1"

from .core import jenks_breaks
from .core import _jenks_matrices
from .core import JenksNaturalBreaks


__all__ = ['jenks_breaks', '_jenks_matrices', 'JenksNaturalBreaks']



# -*- coding: utf-8 -*-
import numpy as np
from collections.abc import Iterable
from . import jenks


class JenksNaturalBreaks:
    """
    A class that can be used to classify a sequence of numbers into groups (clusters) using Fisher-Jenks natural breaks.
    """
    def __init__(self, n_classes=6):
        """
        Parameters
        ----------
        n_classes : int
            The number of classes to be generated by the classifier.
        """
        self.n_classes = n_classes

    def __repr__(self):
        return f"JenksNaturalBreaks(n_classes={self.n_classes})"

    def __str__(self):
        return f"JenksNaturalBreaks(n_classes={self.n_classes})"

    def fit(self, x):
        """
        Parameters
        ----------
        x : array-like
            The Iterable sequence of numbers (integer/float) to be classified.

        """
        self.breaks_ = jenks_breaks(x, self.n_classes)
        self.inner_breaks_ = self.breaks_[1:-1]  # because inner_breaks is more
        self.labels_ = self.predict(x)
        self.groups_ = self.group(x)

    def predict(self, x):
        """
        Predicts the class of each element in x.

        Parameters
        ----------
        x : array-like

        Returns
        -------
        list
        """
        if np.size(x) == 1:
            return np.array(self.get_label_(x, idx=0))
        else:
            labels_ = []
            for val in x:
                label_ = self.get_label_(val, idx=0)
                labels_.append(label_)
            return np.array(labels_)
        
    def group(self, x):
        """
        Groups the elements in x into groups according to the classifier.

        Parameters
        ----------
        x : array-like
            The Iterable sequence of numbers (integer/float) to be classified.

        Returns
        -------
        list
            The list of groups that contains the values of x.
        """
        x = np.array(x)
        groups_ = [x[x <= self.inner_breaks_[0]]]
        for idx in range(len(self.inner_breaks_))[:-1]:
            groups_.append(x[(x > self.inner_breaks_[idx]) * (x <= self.inner_breaks_[idx+1])])  
        groups_.append(x[x > self.inner_breaks_[-1]])
        return groups_
    
    def goodness_of_variance_fit(self, x):
        """
        Parameters
        ----------
        x : array-like

        Returns
        -------
        float
            The goodness of variance fit.
        """
        x = np.array(x)
        array_mean = np.mean(x)
        sdam = sum([(value - array_mean)**2 for value in x])
        sdcm = 0
        for group in self.groups_:
            group_mean = np.mean(group)
            sdcm += sum([(value - group_mean)**2 for value in group])
        gvf = (sdam - sdcm)/sdam
        return gvf

    def get_label_(self, val, idx=0):
        """
        Compute the group label of the given value.

        Parameters
        ----------
        val : float
            The value to be classified.
        idx : int, optional

        Returns
        -------
        int : The label of the value.
        """
        try:
            if val <= self.inner_breaks_[idx]:
                return idx
            else:
                idx = self.get_label_(val, idx+1)
                return idx
        except:
            return len(self.inner_breaks_)


def validate_input(values, n_classes):
    # Check input so that we have an Iterable sequence of numbers
    if not isinstance(values, Iterable) or isinstance(values, (str, bytes)):
        raise TypeError("A sequence of numbers is expected")

    # Number of classes have to be an integer
    if isinstance(n_classes, float) and int(n_classes) == n_classes:
        n_classes = int(n_classes)
    if not isinstance(n_classes, int):
        raise TypeError(
            "Number of class have to be a positive integer: "
            "expected an instance of 'int' but found {}"
            .format(type(n_classes)))

    # Check that numpy arrays are 1-dimensional
    if isinstance(values, np.ndarray):
        if len(values.shape) != 1:
            raise ValueError("A 1D array is expected")

    # Check that all values are finite (and can be cast safely to double later)
    try:
        if not np.isfinite(values).all():
            raise ValueError("All values have to be finite")
    except TypeError as e:
        import sys
        _, value, traceback = sys.exc_info()
        raise TypeError(
            "The sequence of values contains values that will overflow C double capacities - %s"
            .format(value)
        ).with_traceback(traceback)

    # Check that the number of classes is lesser than or equal to the length of unique values
    # and greater than or equal to 1
    if n_classes > len(np.unique(values)) or n_classes < 1:
        raise ValueError("Number of class have to be an integer greater than or equal to 1 and "
                         "smaller than or equal to the number of unique values to use")

    return n_classes

def jenks_breaks(values, n_classes):
    """
    Compute natural breaks (Fisher-Jenks algorithm) on a sequence of `values`,
    given `n_classes`, the number of desired class.

    Parameters
    ----------
    values : array-like
        The Iterable sequence of numbers (integer/float) to be used.
    n_classes : int
        The desired number of class. Have to be lesser than or equal
        to the length of `values` and greater than or equal to 1.

    Returns
    -------
    breaks : list of floats
        The computed break values, including minimum and maximum, in order
        to have all the bounds for building `n_classes` classes,
        so the returned list has a length of `n_classes` + 1.


    Examples
    --------
    Using n_classes = 3, expecting 4 break values , including min and max :

    >>> jenks_breaks(
            [1.3, 7.1, 7.3, 2.3, 3.9, 4.1, 7.8, 1.2, 4.3, 7.3, 5.0, 4.3],
            n_classes = 3)  # Should output [1.2, 2.3, 5.0, 7.8]

    """
    n_classes = validate_input(values, n_classes)
    return jenks._jenks_breaks(values, n_classes)


def _jenks_matrices(values, n_classes, testing_algo=False):
    """
    Returns the intermediate matrices (lower_class_limits and variance combinations)
    that are created when computing natural breaks (Fisher-Jenks algorithm) on a sequence of `values`,
    given `n_classes`, the number of desired class.
    This is only used for testing and debugging purpose, it is not part of the supported API
    and might disappear at any time.

    Parameters
    ----------
    values : array-like
        The Iterable sequence of numbers (integer/float) to be used.
    n_classes : int
        The desired number of class. Have to be lesser than or equal
        to the length of `values` and greater than or equal to 1.
    testing_algo : bool
        If False, use the 'normal' version of the algorithm to compute the
        matrices, if True, use the version of the algorithm that mimic
        Fortran implementation by using the same indexing.

    Returns
    -------
    dict : The computed matrices, as numpy arrays,
           in a dict with keys 'lower_class_limits' and 'variance_combinations'.
    """
    n_classes = validate_input(values, n_classes)

    if testing_algo not in (True, False):
        raise ValueError('testing_algo parameters have to be either True or False')

    return jenks._jenks_matrices(values, n_classes, testing_algo)



# -*- coding: utf-8 -*-
# cython: language_level=3
from libc.stdlib cimport malloc, free
import numpy as np

cdef extern from "_jenks.c":
    void *JenksMatrices(double *variance_combinations, int *lower_class_limits, double *values, unsigned int n_classes, unsigned int length_array)
    void *JenksMatricesTest(double *variance_combinations, int *lower_class_limits, double *values, unsigned int n_classes, unsigned int length_array)


cpdef list _jenks_breaks(values, n_classes):
    cdef:
        list breaks = []
        unsigned int _n_classes = <unsigned int>int(n_classes) # How many classes (clusters)
        unsigned int _n_breaks = <unsigned int>_n_classes + 1 # How many class boundaries
        unsigned int len_values = <unsigned int>len(values) # How many values to classify
        unsigned int i
        int *_lower_class_limits = <int *> malloc((len_values * _n_classes) * sizeof(int))
        double *_variance_combinations = <double *> malloc((len_values * _n_classes) * sizeof(double))
        unsigned int m, k, jj # Indexes for fetching break values from _lower_class_limits matrix
        sorted_values = np.sort(values) # Sort the values to classify
        double[:] _sorted_values = sorted_values.astype(np.float64) # Convert to double array

    # Fill the lower_class_limits and variance_combinations matrices
    JenksMatrices(_variance_combinations, _lower_class_limits, &_sorted_values[0], _n_classes, len_values)

    # Fill the breaks array with minimum value of the target array
    for i in range(_n_breaks):
        breaks.append(sorted_values[0])

    # Last break value is the maximum of the array
    breaks[_n_classes] = sorted_values[-1]

    # Read the class limits from the lower_class_limits matrix
    m = len_values
    for j in range(1, _n_classes):
        jj = _n_classes - j + 1
        # Note that the indexes contained in the _lower_class_limits matrix are 1-based,
        # so we have to subtract 1 to get the correct index
        breaks[jj - 1] = sorted_values[_lower_class_limits[(m - 1) * _n_classes + (jj - 1)] - 2]
        m = _lower_class_limits[(m - 1) * _n_classes + (jj - 1)] - 1

    # Free memory we manually allocated
    free(<void *>_lower_class_limits)
    free(<void *>_variance_combinations)

    return breaks

cpdef dict _jenks_matrices(values, n_classes, testing_algo=False):
    cdef:
        unsigned int _n_classes = <unsigned int>int(n_classes) # How many classes (clusters)
        unsigned int len_values = <unsigned int>len(values) # How many values to classify
        unsigned int i
        int *_lower_class_limits = <int *> malloc((len_values * _n_classes) * sizeof(int))
        double *_variance_combinations = <double *> malloc((len_values * _n_classes) * sizeof(double))
        unsigned int m, k, jj # Indexes for fetching break values from _lower_class_limits matrix
        list lower_class_limits = []
        list variance_combinations = []
        sorted_values = np.sort(values) # Sort the values to classify
        double[:] _sorted_values = sorted_values.astype(np.float64) # Convert to double array

    # Fill the lower_class_limits and variance_combinations matrices
    if not testing_algo:
        JenksMatrices(_variance_combinations, _lower_class_limits, &_sorted_values[0], _n_classes, len_values)
    else:
        JenksMatricesTest(_variance_combinations, _lower_class_limits, &_sorted_values[0], _n_classes, len_values)

    # Reconstruct the lower class limits matrix and the variance combinations matrix
    # from the 1D arrays that we use in C code
    for i in range(len_values):
        lower_class_limits.append([])
        variance_combinations.append([])
        for j in range(_n_classes):
            lower_class_limits[i].append(_lower_class_limits[i * _n_classes + j])
            variance_combinations[i].append(_variance_combinations[i * _n_classes + j])

    # Free memory we manually allocated
    free(<void *>_lower_class_limits)
    free(<void *>_variance_combinations)

    return {
        "lower_class_limits": np.array(lower_class_limits),
        "variance_combinations": np.array(variance_combinations),
    }

